﻿/*
 * font.c
 * Модуль для вывода текстовой информации, работает совместно с wg12232.h
 * 
 * Author: Погребняк Дмитрий, г. Самара, 2013
 *
 * Помещённый здесь код, включая шрифты, является свободным. Т.е. допускается его свободное использование для любых целей, включая коммерческие, при условии указания ссылки на автора (Погребняк Дмитрий, http://aterlux.ru/).
 * Code and fonts included here are free to use. It can be used for free for any purpose, including commercial use, as long as link to author (Pogrebnyak Dmitry, http://aterlux.ru/) is included.
 */ 

#include <avr/io.h>
#include <avr/pgmspace.h>
#include "font.h"
#include "Tools.h"
#include "wg12232.h"

/*
  Формат шрифта:
 
  Позиция   Длина     Описание
  0         1         Флаги: 
                        бит 7: 1, если старший бит вверху, младший внизу 0 - младший вверху, старший внизу; 
                        бит 1: добавлять пустую колонку перед символом 
                        бит 0: добавлять пустую колонку после символа
  1         1         Вертикальные габариты: 0xtb. Высота шрифта = b - t + 1:
                        биты 6..4: t - первая строка с которой начинается строка шрифта (у полноразмерных шрифтов - 0)
                        биты 2..0: b - последняя строка, на которой заканчивается строка шрифта (у полноразмерных - 7)
  2         1         Первый код символа, присутствующий в шрифте S
  3         1         Последний код символа, присутствующий в шрифте E (должно быть не меньше S)
  4         (E-S+1)*2 Таблица размещения символов и их ширин (сначала младший байт) c S по E:
                      биты 0..11: offset[i] - позиция от начала массива где начинаются данные символа
                      биты 12..15: width[i] - ширина символа 
  offset[i] width[i]  Данные символов
                        Символы кодируются вертикальными линиями, слева направо. Один байт - одна вертикальная линия. Порядок бит определяется значением соответствующего флага
                        Разные символы могут использовать одни и те же позиции данных, вплоть до возможности их пересечения

  Font structure:
  Position  Length    Description
  0         1         Flags:
                        bit 7: 1 - Most Significant Bit on top; 0 - Least Significant Bit on top
                        bit 1: if set, add empty column before every character
                        bit 0: if set, add empty column after every character
  1         1         Vertical size: Height of a font = b - t + 1
                        bits 6..4: t - number of first row (from top)
                        bits 2..0: b - last pixel line
  2         1         S: Code of the first character in a font
  3         1         E: Code of the last character in a font (should be >= S)
  4         (E-S+1)*2 The table of character data offsets and character widths, two bytes (LSB first) for every character from S to E, inclusive:
                        bits 0..11: offset[i] - offset of a character data, from font beginning 
                        bits 12..15: width[i] - the width of a character
  offset[i] width[i]  Font characters data
                        The characters encoded by vertical lines, each byte for each vertical line, from left to right. Bit order determined by bit 7 of Flags
                        Different character may share same character data, even overlap
*/

// Мелкий шрифт, 6 пикселей в высоту 
// Micro font, 6 pixel height
const uint8_t font_micro[] PROGMEM = {
  0x81, 0x05, 0x18, 0xC1, 0x58, 0x51, 0x5C, 0x51, 0x60, 0x41, 0x63, 0x41, 0x67, 0x31, 0x6A, 0x71, 0x71, 0x51, 0x76, 0x51, 0x7B, 0x11, 0x7C, 0x11, 0x7D, 0x31, 0x80, 0x51, 0x85, 0x41, 0x89, 0x31,
  0x8C, 0x51, 0x7D, 0x11, 0x91, 0x21, 0x92, 0x21, 0x94, 0x51, 0x99, 0x51, 0x67, 0x11, 0x9C, 0x31, 0x68, 0x11, 0x9F, 0x31, 0x91, 0x31, 0x5A, 0x11, 0xA2, 0x31, 0xA5, 0x21, 0xA7, 0x31, 0xAA, 0x31,
  0xAD, 0x31, 0xB0, 0x31, 0xB3, 0x31, 0xB6, 0x31, 0x85, 0x11, 0xB9, 0x11, 0xBA, 0x31, 0xBD, 0x31, 0xC0, 0x31, 0xC3, 0x31, 0xC6, 0x51, 0xCA, 0x31, 0xCD, 0x31, 0xD0, 0x31, 0xD3, 0x31, 0xD6, 0x31,
  0xD9, 0x31, 0xDC, 0x31, 0xDF, 0x31, 0x5A, 0x11, 0xE2, 0x31, 0xE5, 0x31, 0xE8, 0x31, 0xEB, 0x51, 0xEF, 0x41, 0x91, 0x31, 0xF2, 0x31, 0xF5, 0x31, 0xF8, 0x31, 0xFB, 0x31, 0xFE, 0x31, 0x01, 0x32,
  0x04, 0x32, 0x07, 0x52, 0x0C, 0x32, 0x0F, 0x32, 0x12, 0x32, 0xD3, 0x21, 0x15, 0x32, 0xD2, 0x21, 0x18, 0x32, 0x1B, 0x32, 0x19, 0x22, 0x1E, 0x42, 0x22, 0x32, 0x25, 0x22, 0x27, 0x32, 0x2A, 0x32,
  0x2D, 0x22, 0x2F, 0x32, 0x32, 0x32, 0x86, 0x11, 0x35, 0x32, 0x38, 0x32, 0x5A, 0x11, 0x3B, 0x52, 0x3B, 0x32, 0x40, 0x32, 0x43, 0x32, 0x45, 0x42, 0x49, 0x32, 0x4C, 0x32, 0x4F, 0x22, 0x51, 0x32,
  0x54, 0x32, 0x56, 0x52, 0x44, 0x32, 0x5B, 0x32, 0x5E, 0x22, 0x0D, 0x22, 0x5A, 0x11, 0x0C, 0x22, 0x60, 0x42, 0x64, 0x52, 0xCA, 0x31, 0x69, 0x32, 0xCD, 0x31, 0x6C, 0x32, 0x6F, 0x52, 0xD6, 0x31,
  0x74, 0x52, 0x78, 0x32, 0x7B, 0x42, 0x7F, 0x42, 0xE5, 0x31, 0x83, 0x42, 0xEB, 0x51, 0xDF, 0x31, 0x91, 0x31, 0x86, 0x32, 0xF2, 0x31, 0xD0, 0x31, 0xFE, 0x31, 0x89, 0x32, 0x8C, 0x52, 0x0C, 0x32,
  0x91, 0x42, 0x95, 0x32, 0x97, 0x52, 0x9B, 0x62, 0xA1, 0x42, 0xA2, 0x42, 0x22, 0x32, 0xA6, 0x32, 0xA9, 0x42, 0xAD, 0x32, 0x1E, 0x42, 0xAD, 0x31, 0xB0, 0x32, 0x3B, 0x22, 0xB3, 0x32, 0x2A, 0x32,
  0xB6, 0x52, 0xBA, 0x32, 0xBD, 0x42, 0x7F, 0x42, 0xC0, 0x32, 0xC3, 0x42, 0xC6, 0x52, 0xCA, 0x32, 0x40, 0x32, 0xCC, 0x32, 0x43, 0x32, 0x25, 0x22, 0xCD, 0x32, 0x5B, 0x32, 0xD0, 0x52, 0x44, 0x32,
  0xD5, 0x42, 0xD9, 0x32, 0xDB, 0x52, 0xDF, 0x52, 0xE4, 0x42, 0xE5, 0x42, 0xE5, 0x32, 0xE9, 0x32, 0xEC, 0x42, 0xF0, 0x32, 0xF3, 0x32, 0xF6, 0x32, 0x20, 0x40, 0xF8, 0x40, 0x20, 0x10, 0xF8, 0x10,
  0x20, 0xA8, 0x70, 0x20, 0x70, 0xA8, 0x20, 0x18, 0x08, 0x08, 0x20, 0x70, 0xA8, 0x20, 0xA8, 0x70, 0x20, 0x08, 0x38, 0xF8, 0x38, 0x08, 0x80, 0xE0, 0xF8, 0xE0, 0x80, 0x00, 0xF4, 0xC0, 0x00, 0xC0,
  0x50, 0xF8, 0x50, 0xF8, 0x50, 0x48, 0xB8, 0xE8, 0x90, 0x98, 0x20, 0xC8, 0x50, 0xA8, 0x68, 0x10, 0x28, 0x70, 0x88, 0x70, 0x50, 0x20, 0xF8, 0x20, 0x50, 0x20, 0x20, 0xF8, 0x20, 0x20, 0x20, 0x18,
  0x20, 0xC0, 0x48, 0x98, 0x68, 0xA8, 0x50, 0x30, 0x50, 0xF8, 0xE8, 0xA8, 0x90, 0x70, 0xA8, 0x10, 0x80, 0xB8, 0xC0, 0x50, 0xA8, 0x50, 0x40, 0xA8, 0x70, 0x58, 0x20, 0x50, 0x88, 0x50, 0x50, 0x50,
  0x88, 0x50, 0x20, 0x80, 0xA8, 0x40, 0x70, 0xA8, 0xD8, 0xB0, 0x78, 0x90, 0x78, 0xF8, 0xA8, 0x50, 0x70, 0x88, 0x88, 0xF8, 0x88, 0x70, 0xF8, 0xA8, 0x88, 0xF8, 0xA0, 0x80, 0x70, 0x88, 0xB8, 0xF8,
  0x20, 0xF8, 0x10, 0x08, 0xF0, 0xF8, 0x60, 0x98, 0xF8, 0x08, 0x08, 0xF8, 0x60, 0x30, 0x60, 0xF8, 0x40, 0x20, 0xF8, 0xA0, 0x40, 0x70, 0x98, 0x78, 0xF8, 0xA0, 0x58, 0x48, 0xA8, 0x90, 0x80, 0xF8,
  0x80, 0xF0, 0x08, 0xF8, 0xF0, 0x08, 0xF0, 0xF8, 0x30, 0x60, 0x30, 0xF8, 0xD8, 0x20, 0xD8, 0xC0, 0x38, 0xC0, 0x98, 0xA8, 0xC8, 0xC0, 0x20, 0x18, 0x40, 0x80, 0x40, 0x08, 0x08, 0x08, 0x50, 0x68,
  0x30, 0x08, 0xF8, 0x28, 0x10, 0x30, 0x48, 0x10, 0x28, 0xF8, 0x30, 0x58, 0x28, 0x78, 0xA0, 0x20, 0x54, 0x78, 0xF8, 0x20, 0x18, 0x10, 0x08, 0xB0, 0xF8, 0x10, 0x28, 0x78, 0x40, 0x38, 0x40, 0x38,
  0x30, 0x48, 0x30, 0x7C, 0x48, 0x30, 0x48, 0x7C, 0x08, 0x78, 0x20, 0x40, 0x28, 0x68, 0x50, 0xF0, 0x48, 0x70, 0x08, 0x78, 0x70, 0x08, 0x70, 0x08, 0x30, 0x08, 0x70, 0x60, 0x14, 0x78, 0x58, 0x68,
  0x40, 0x80, 0x40, 0x80, 0x30, 0x50, 0x90, 0x50, 0x30, 0xF8, 0xA8, 0x90, 0xF8, 0x80, 0x80, 0x08, 0x78, 0x88, 0xF8, 0x08, 0x88, 0x50, 0xF8, 0x50, 0x88, 0xA8, 0x50, 0xF8, 0x10, 0x20, 0xF8, 0x78,
  0x90, 0xA0, 0x78, 0x08, 0x70, 0x80, 0xF8, 0x80, 0xF8, 0xC8, 0x28, 0xF0, 0x20, 0x50, 0xF8, 0x50, 0x20, 0xF8, 0x08, 0xF8, 0x0C, 0xC0, 0x20, 0xF8, 0x08, 0x78, 0x08, 0xF8, 0x08, 0xF8, 0x08, 0xF8,
  0x0C, 0x80, 0xF8, 0x28, 0x10, 0xF8, 0x88, 0xA8, 0x70, 0xF8, 0x70, 0x88, 0x70, 0x68, 0x90, 0xF8, 0x78, 0x68, 0x10, 0x90, 0xA8, 0x70, 0x48, 0x30, 0x78, 0x30, 0x48, 0x68, 0x50, 0x78, 0x10, 0x20,
  0x78, 0x20, 0x58, 0x08, 0x30, 0x40, 0x78, 0x20, 0x10, 0x20, 0x78, 0x20, 0x78, 0x40, 0x78, 0x40, 0x20, 0x50, 0x78, 0x50, 0x20, 0x78, 0x08, 0x78, 0x0C, 0x60, 0x10, 0x78, 0x08, 0x38, 0x08, 0x78,
  0x08, 0x38, 0x08, 0x7C, 0x40, 0x78, 0x28, 0x10, 0x78, 0x48, 0x68, 0x30, 0x78, 0x30, 0x48, 0x30, 0x28, 0x50, 0x78, 0xF8, 0x68, 0xC8, 0xB0, 0x58, 0xA8
};

// Обычный шрифт, 8 пикселей в высоту
// Normal font, 8 pixel height
const uint8_t font_normal[] PROGMEM = {
  0x81, 0x07, 0x18, 0xC1, 0x58, 0x51, 0x5D, 0x51, 0x62, 0x51, 0x66, 0x51, 0x6B, 0x61, 0x71, 0x71, 0x78, 0x71, 0x7F, 0x71, 0x86, 0x11, 0x87, 0x11, 0x88, 0x31, 0x8B, 0x51, 0x90, 0x51, 0x95, 0x51,
  0x9A, 0x51, 0x9F, 0x21, 0xA1, 0x21, 0xA2, 0x21, 0xA4, 0x51, 0xA9, 0x51, 0xAE, 0x21, 0xB0, 0x51, 0x5E, 0x11, 0xB5, 0x31, 0xB8, 0x51, 0xBD, 0x31, 0xC0, 0x51, 0xC5, 0x51, 0xCA, 0x51, 0xCF, 0x51,
  0xD4, 0x51, 0xD9, 0x51, 0xDE, 0x51, 0xE3, 0x51, 0x8B, 0x11, 0xE8, 0x21, 0xEA, 0x31, 0xED, 0x51, 0xF2, 0x31, 0xF5, 0x51, 0xFA, 0x51, 0xFF, 0x51, 0x04, 0x52, 0x09, 0x52, 0x0E, 0x52, 0x13, 0x52,
  0x18, 0x52, 0x1D, 0x52, 0x22, 0x52, 0x27, 0x32, 0x2A, 0x42, 0x2E, 0x52, 0x33, 0x52, 0x38, 0x52, 0x3C, 0x52, 0xB8, 0x51, 0x40, 0x52, 0x45, 0x52, 0x4A, 0x52, 0x4F, 0x52, 0x54, 0x52, 0x59, 0x52,
  0x5E, 0x52, 0x63, 0x52, 0x68, 0x52, 0x6D, 0x52, 0x72, 0x52, 0x0E, 0x32, 0x77, 0x32, 0x7A, 0x32, 0x7D, 0x52, 0x82, 0x52, 0x5B, 0x21, 0x87, 0x52, 0x8C, 0x52, 0x90, 0x52, 0x95, 0x52, 0x9A, 0x52,
  0x9E, 0x42, 0xA2, 0x52, 0xA7, 0x52, 0xAC, 0x32, 0xAE, 0x32, 0xB1, 0x52, 0xB6, 0x32, 0xB9, 0x52, 0xBE, 0x52, 0xC3, 0x52, 0xC8, 0x52, 0xCC, 0x62, 0xD2, 0x52, 0xD7, 0x52, 0xDC, 0x42, 0xE0, 0x52,
  0xE5, 0x52, 0xEA, 0x52, 0xEF, 0x52, 0xF4, 0x52, 0xF9, 0x52, 0xFE, 0x32, 0x01, 0x13, 0x02, 0x33, 0x04, 0x53, 0x09, 0x53, 0xFF, 0x51, 0x0E, 0x53, 0x04, 0x52, 0x13, 0x53, 0x18, 0x53, 0x13, 0x52,
  0x1D, 0x53, 0x22, 0x53, 0x27, 0x53, 0x2C, 0x53, 0x2E, 0x52, 0x31, 0x53, 0x38, 0x52, 0x22, 0x52, 0xB8, 0x51, 0x35, 0x53, 0x40, 0x52, 0x09, 0x52, 0x54, 0x52, 0x3A, 0x53, 0x3F, 0x53, 0x68, 0x52,
  0x44, 0x53, 0x49, 0x53, 0x4D, 0x53, 0x51, 0x63, 0x57, 0x53, 0x58, 0x53, 0x5C, 0x53, 0x61, 0x53, 0x66, 0x53, 0x6B, 0x53, 0x87, 0x52, 0xD4, 0x51, 0x70, 0x53, 0x75, 0x53, 0x7A, 0x53, 0x9A, 0x52,
  0x7F, 0x53, 0x84, 0x53, 0x89, 0x53, 0x8D, 0x53, 0x91, 0x53, 0x96, 0x53, 0x9A, 0x53, 0x9E, 0x53, 0xC3, 0x52, 0xA2, 0x53, 0xC8, 0x52, 0x90, 0x52, 0xA4, 0x53, 0xF4, 0x52, 0xA9, 0x53, 0xEF, 0x52,
  0xAE, 0x53, 0xB3, 0x53, 0xB7, 0x53, 0xB7, 0x63, 0xBD, 0x53, 0xBE, 0x53, 0xC2, 0x53, 0xC7, 0x53, 0xCC, 0x53, 0xD1, 0x53, 0xD6, 0x53, 0xDB, 0x53, 0x20, 0x40, 0xFE, 0x40, 0x20, 0x08, 0x04, 0xFE,
  0x04, 0x08, 0x10, 0x10, 0x54, 0x38, 0x10, 0x38, 0x54, 0x10, 0x10, 0x1C, 0x1C, 0x04, 0x04, 0x04, 0x04, 0x10, 0x38, 0x54, 0x10, 0x54, 0x38, 0x10, 0x02, 0x0E, 0x3E, 0xFE, 0x3E, 0x0E, 0x02, 0x80,
  0xE0, 0xF8, 0xFE, 0xF8, 0xE0, 0x80, 0x00, 0xFA, 0xC0, 0x00, 0xC0, 0x24, 0x7E, 0x24, 0x7E, 0x24, 0x64, 0x92, 0xFE, 0x92, 0x4C, 0xC2, 0xCC, 0x10, 0x66, 0x86, 0x0C, 0x72, 0x92, 0x6C, 0x0A, 0x40,
  0x80, 0x7C, 0x82, 0x7C, 0x28, 0x10, 0x7C, 0x10, 0x28, 0x10, 0x10, 0x7C, 0x10, 0x10, 0x02, 0x04, 0x10, 0x10, 0x10, 0x10, 0x10, 0x06, 0x18, 0x60, 0x7C, 0x82, 0x82, 0x82, 0x7C, 0x42, 0xFE, 0x02,
  0x42, 0x86, 0x8A, 0x92, 0x62, 0x84, 0x82, 0x92, 0xB2, 0xCC, 0x18, 0x28, 0x48, 0xFE, 0x08, 0xE4, 0xA2, 0xA2, 0xA2, 0x9C, 0x7C, 0xA2, 0xA2, 0xA2, 0x1C, 0x80, 0x80, 0x8E, 0xB0, 0xC0, 0x6C, 0x92,
  0x92, 0x92, 0x6C, 0x60, 0x92, 0x92, 0x92, 0x7C, 0x02, 0x24, 0x10, 0x28, 0x44, 0x28, 0x28, 0x28, 0x28, 0x28, 0x44, 0x28, 0x10, 0x40, 0x80, 0x8A, 0x90, 0x60, 0x38, 0x44, 0x92, 0xAA, 0x7A, 0x7E,
  0x88, 0x88, 0x88, 0x7E, 0xFE, 0x92, 0x92, 0x72, 0x0C, 0x7C, 0x82, 0x82, 0x82, 0x44, 0xFE, 0x82, 0x82, 0x44, 0x38, 0xFE, 0x92, 0x92, 0x92, 0x82, 0xFE, 0x90, 0x90, 0x90, 0x80, 0x7C, 0x82, 0x82,
  0x8A, 0x4C, 0xFE, 0x10, 0x10, 0x10, 0xFE, 0x82, 0xFE, 0x82, 0x04, 0x02, 0x02, 0xFC, 0xFE, 0x10, 0x30, 0x48, 0x86, 0xFE, 0x02, 0x02, 0x02, 0x02, 0xFE, 0x60, 0x18, 0x60, 0xFE, 0x40, 0x30, 0x08,
  0xFE, 0x90, 0x90, 0x90, 0x60, 0x7C, 0x82, 0x82, 0x84, 0x7A, 0xFE, 0x90, 0x98, 0x94, 0x62, 0x64, 0x92, 0x92, 0x92, 0x4C, 0x80, 0x80, 0xFE, 0x80, 0x80, 0xFC, 0x02, 0x02, 0x02, 0xFC, 0xE0, 0x18,
  0x06, 0x18, 0xE0, 0xFC, 0x06, 0x08, 0x06, 0xFC, 0xC6, 0x28, 0x10, 0x28, 0xC6, 0xE0, 0x10, 0x0E, 0x10, 0xE0, 0x86, 0x8A, 0x92, 0xA2, 0xC2, 0x60, 0x18, 0x06, 0x82, 0x82, 0xFE, 0x20, 0x40, 0x80,
  0x40, 0x20, 0x02, 0x02, 0x02, 0x02, 0x02, 0x04, 0x2A, 0x2A, 0x1C, 0x02, 0xFE, 0x24, 0x22, 0x22, 0x1C, 0x22, 0x22, 0x22, 0x14, 0x1C, 0x22, 0x22, 0x22, 0xFE, 0x1C, 0x2A, 0x2A, 0x2A, 0x10, 0x7E,
  0x90, 0x80, 0x18, 0x25, 0x25, 0x15, 0x3E, 0xFE, 0x20, 0x20, 0x20, 0x1E, 0x22, 0xBE, 0x02, 0x21, 0xBE, 0xFE, 0x08, 0x18, 0x24, 0x02, 0x82, 0xFE, 0x02, 0x3E, 0x20, 0x1E, 0x20, 0x1E, 0x3E, 0x10,
  0x20, 0x20, 0x1E, 0x1C, 0x22, 0x22, 0x22, 0x1C, 0x3F, 0x24, 0x24, 0x24, 0x18, 0x24, 0x24, 0x24, 0x3F, 0x02, 0x3E, 0x10, 0x20, 0x20, 0x20, 0x12, 0x2A, 0x2A, 0x2A, 0x24, 0x20, 0xFC, 0x22, 0x02,
  0x3C, 0x02, 0x02, 0x02, 0x3C, 0x30, 0x0C, 0x02, 0x0C, 0x30, 0x38, 0x06, 0x08, 0x06, 0x38, 0x22, 0x14, 0x08, 0x14, 0x22, 0x30, 0x09, 0x06, 0x0C, 0x30, 0x22, 0x26, 0x2A, 0x32, 0x22, 0x10, 0x6C,
  0x82, 0xFF, 0x82, 0x6C, 0x10, 0x20, 0x10, 0x08, 0x10, 0x3E, 0x42, 0x82, 0x42, 0x3E, 0xFE, 0x92, 0x92, 0x92, 0x0C, 0xFE, 0x80, 0x80, 0x80, 0x80, 0x03, 0x7E, 0x82, 0xFE, 0x03, 0xC6, 0x28, 0xFE,
  0x28, 0xC6, 0x44, 0x82, 0x92, 0x92, 0x6C, 0xFE, 0x04, 0x08, 0x10, 0xFE, 0x7E, 0x84, 0x88, 0x90, 0x7E, 0x02, 0x7C, 0x80, 0x80, 0xFE, 0x80, 0x80, 0x80, 0xFE, 0xC0, 0x32, 0x0C, 0x30, 0xC0, 0x30,
  0x48, 0xFE, 0x48, 0x30, 0xFE, 0x02, 0x02, 0xFE, 0x03, 0xE0, 0x10, 0x10, 0x10, 0xFE, 0x02, 0x7E, 0x02, 0xFE, 0x02, 0xFE, 0x02, 0xFE, 0x03, 0x80, 0xFE, 0x12, 0x12, 0x0C, 0xFE, 0x12, 0x12, 0x12,
  0x0C, 0x44, 0x82, 0x92, 0x92, 0x7C, 0xFE, 0x10, 0x7C, 0x82, 0x7C, 0x62, 0x94, 0x98, 0x90, 0xFE, 0x3E, 0x2A, 0x2A, 0x1A, 0x04, 0x3E, 0x20, 0x20, 0x20, 0x20, 0x03, 0x1E, 0x22, 0x3E, 0x03, 0x36,
  0x08, 0x3E, 0x08, 0x36, 0x14, 0x22, 0x22, 0x2A, 0x14, 0x3E, 0x04, 0x08, 0x10, 0x3E, 0x84, 0x48, 0x90, 0x3E, 0x08, 0x18, 0x28, 0x06, 0x02, 0x3C, 0x20, 0x20, 0x3E, 0x10, 0x08, 0x10, 0x3E, 0x08,
  0x08, 0x08, 0x3E, 0x20, 0x20, 0x20, 0x3E, 0x20, 0x20, 0x08, 0x14, 0x3E, 0x14, 0x08, 0x3E, 0x02, 0x02, 0x3E, 0x03, 0x30, 0x08, 0x08, 0x08, 0x3E, 0x02, 0x1E, 0x02, 0x3E, 0x03, 0x20, 0x3E, 0x0A,
  0x0A, 0x04, 0x3E, 0x0A, 0x0A, 0x0A, 0x04, 0x14, 0x22, 0x2A, 0x2A, 0x1C, 0x3E, 0x08, 0x1C, 0x22, 0x1C, 0x12, 0x2C, 0x28, 0x28, 0x3E, 0x7E, 0xD2, 0x52, 0xD2, 0x42, 0x1C, 0xAA, 0x2A, 0xAA, 0x10
};

// Жирный шрифт, 8 пикселей в высоту
// Fat font, 8 pixel height
const uint8_t font_bold[] PROGMEM = {
  0x81, 0x07, 0x18, 0xC1, 0x58, 0x51, 0x5D, 0x51, 0x62, 0x71, 0x68, 0x71, 0x6F, 0x61, 0x75, 0x71, 0x7C, 0x71, 0x83, 0x71, 0x8A, 0x11, 0x8B, 0x41, 0x8F, 0x51, 0x94, 0x71, 0x9B, 0x71, 0xA2, 0x71,
  0xA9, 0x71, 0x8F, 0x21, 0xB0, 0x31, 0xB2, 0x31, 0xB5, 0x51, 0xBA, 0x51, 0xBF, 0x21, 0xC1, 0x51, 0xC6, 0x21, 0xC8, 0x71, 0xCF, 0x71, 0xD6, 0x41, 0xDA, 0x71, 0xE1, 0x71, 0xE8, 0x71, 0xEF, 0x71,
  0xF6, 0x71, 0xFD, 0x71, 0x04, 0x72, 0x0B, 0x72, 0x12, 0x22, 0x14, 0x22, 0x16, 0x52, 0x1B, 0x52, 0x20, 0x52, 0x25, 0x72, 0x2C, 0x72, 0x33, 0x72, 0x3A, 0x72, 0x41, 0x72, 0x48, 0x72, 0x4F, 0x72,
  0x56, 0x72, 0x5D, 0x72, 0x64, 0x72, 0x6B, 0x42, 0x6F, 0x72, 0x76, 0x72, 0x7D, 0x72, 0x84, 0x72, 0x89, 0x72, 0x90, 0x72, 0x97, 0x72, 0x9E, 0x72, 0xA5, 0x72, 0xAC, 0x72, 0xB3, 0x62, 0xB9, 0x72,
  0xC0, 0x72, 0xC7, 0x72, 0xCE, 0x72, 0xD5, 0x62, 0xDB, 0x72, 0x48, 0x42, 0xE2, 0x72, 0xE9, 0x42, 0xED, 0x52, 0xF2, 0x52, 0xF7, 0x22, 0xF9, 0x72, 0x00, 0x73, 0x06, 0x73, 0x0D, 0x73, 0x14, 0x73,
  0x1A, 0x53, 0x1F, 0x73, 0x26, 0x73, 0x2D, 0x43, 0x30, 0x63, 0x36, 0x73, 0x3D, 0x43, 0x41, 0x73, 0x48, 0x73, 0x4F, 0x73, 0x56, 0x73, 0x5C, 0x73, 0x63, 0x73, 0x6A, 0x73, 0x71, 0x53, 0x76, 0x73,
  0x7D, 0x73, 0x84, 0x73, 0x8B, 0x73, 0x92, 0x73, 0x99, 0x73, 0xA0, 0x43, 0x95, 0x21, 0xA3, 0x43, 0xA7, 0x43, 0xAB, 0x73, 0xB1, 0x73, 0xB8, 0x73, 0x3A, 0x72, 0xBF, 0x73, 0xC6, 0x73, 0x4F, 0x72,
  0xCD, 0x73, 0xD4, 0x73, 0xDB, 0x73, 0xE1, 0x73, 0x76, 0x72, 0xE8, 0x73, 0xED, 0x73, 0x64, 0x72, 0x90, 0x72, 0xF2, 0x73, 0x97, 0x72, 0x41, 0x72, 0xB3, 0x62, 0xF9, 0x73, 0x00, 0x64, 0x06, 0x74,
  0x0D, 0x74, 0x14, 0x74, 0x19, 0x74, 0x1E, 0x84, 0x26, 0x74, 0x2D, 0x74, 0x33, 0x74, 0x3A, 0x74, 0x41, 0x74, 0x48, 0x74, 0xF9, 0x72, 0x4F, 0x74, 0x56, 0x74, 0x5D, 0x64, 0x63, 0x74, 0x6A, 0x74,
  0x71, 0x74, 0x78, 0x74, 0x7F, 0x74, 0x84, 0x74, 0x89, 0x74, 0x8F, 0x74, 0x94, 0x74, 0x99, 0x74, 0x4F, 0x73, 0x9E, 0x74, 0xA5, 0x74, 0x06, 0x73, 0xAC, 0x64, 0xB2, 0x74, 0xB9, 0x64, 0x8B, 0x73,
  0xBF, 0x74, 0xC6, 0x74, 0xCB, 0x74, 0xD0, 0x84, 0xD8, 0x74, 0xDF, 0x74, 0xE5, 0x74, 0xEC, 0x74, 0xF3, 0x74, 0xFA, 0x74, 0x01, 0x75, 0x08, 0x75, 0x10, 0x30, 0x7E, 0x30, 0x10, 0x08, 0x0C, 0x7E,
  0x0C, 0x08, 0x10, 0x10, 0x10, 0x10, 0x7C, 0x38, 0x10, 0x38, 0x7C, 0x10, 0x10, 0x10, 0x10, 0x1C, 0x1C, 0x04, 0x04, 0x04, 0x04, 0x08, 0x1C, 0x3E, 0x08, 0x3E, 0x1C, 0x08, 0x02, 0x0E, 0x3E, 0xFE,
  0x3E, 0x0E, 0x02, 0x80, 0xE0, 0xF8, 0xFE, 0xF8, 0xE0, 0x80, 0x00, 0x60, 0xF6, 0xF6, 0x60, 0xA0, 0xC0, 0x00, 0xA0, 0xC0, 0x44, 0xFE, 0xFE, 0x44, 0xFE, 0xFE, 0x44, 0x24, 0x52, 0xFE, 0x52, 0xFE,
  0x52, 0x4C, 0x42, 0xA4, 0x48, 0x10, 0x24, 0x4A, 0x84, 0x0C, 0x72, 0x92, 0x92, 0x6A, 0x0C, 0x14, 0x7C, 0xFE, 0x82, 0xFE, 0x7C, 0x28, 0x10, 0x7C, 0x10, 0x28, 0x10, 0x10, 0x7C, 0x10, 0x10, 0x05,
  0x06, 0x10, 0x10, 0x10, 0x10, 0x10, 0x06, 0x06, 0x02, 0x04, 0x08, 0x10, 0x20, 0x40, 0x80, 0x38, 0x7C, 0x82, 0x82, 0x82, 0x7C, 0x38, 0x42, 0xFE, 0xFE, 0x02, 0x46, 0xCE, 0x8A, 0x9A, 0x92, 0xF2,
  0x66, 0x44, 0xC6, 0x82, 0x92, 0x92, 0xFE, 0x6C, 0x08, 0x18, 0x28, 0x4A, 0xFE, 0xFE, 0x0A, 0xE4, 0xE6, 0xA2, 0xA2, 0xA2, 0xBE, 0x9C, 0x3C, 0x7E, 0xD2, 0x92, 0x92, 0xDE, 0x4C, 0xC0, 0xC0, 0x80,
  0x8E, 0x9E, 0xF0, 0xE0, 0x6C, 0xFE, 0x92, 0x92, 0x92, 0xFE, 0x6C, 0x70, 0xFA, 0x8A, 0x8A, 0x8A, 0xFE, 0x7C, 0x66, 0x66, 0x65, 0x66, 0x10, 0x38, 0x6C, 0xC6, 0x82, 0x28, 0x28, 0x28, 0x28, 0x28,
  0x82, 0xC6, 0x6C, 0x38, 0x10, 0x40, 0xC0, 0x80, 0x8A, 0x9A, 0xF0, 0x60, 0x38, 0x44, 0x92, 0xAA, 0x92, 0x4A, 0x30, 0x3E, 0x7E, 0xC8, 0xC8, 0xC8, 0x7E, 0x3E, 0xFE, 0xFE, 0x92, 0x92, 0xF2, 0x5E,
  0x0C, 0x7C, 0xFE, 0x82, 0x82, 0x82, 0xC6, 0x44, 0xFE, 0xFE, 0x82, 0x82, 0xC6, 0x7C, 0x38, 0xFE, 0xFE, 0x92, 0x92, 0x82, 0xC6, 0xC6, 0xFE, 0xFE, 0x90, 0x90, 0x80, 0xC0, 0xC0, 0x7C, 0xFE, 0x82,
  0x92, 0x92, 0xDE, 0x5C, 0xFE, 0xFE, 0x10, 0x10, 0x10, 0xFE, 0xFE, 0x82, 0xFE, 0xFE, 0x82, 0x0C, 0x0E, 0x02, 0x02, 0x02, 0xFE, 0xFC, 0xFE, 0xFE, 0x10, 0x30, 0x78, 0xCE, 0x86, 0xFE, 0xFE, 0x02,
  0x02, 0x02, 0x06, 0x06, 0xFE, 0xFE, 0x60, 0x18, 0x60, 0xFE, 0xFE, 0x20, 0x10, 0x08, 0xFE, 0xFE, 0x7C, 0xFE, 0x82, 0x82, 0x82, 0xFE, 0x7C, 0xFE, 0xFE, 0x88, 0x88, 0x88, 0xF8, 0x70, 0x7C, 0xFE,
  0x82, 0x8A, 0x86, 0xFC, 0x7A, 0xFE, 0xFE, 0x88, 0x88, 0x8C, 0xFE, 0x72, 0x64, 0xF6, 0x92, 0x92, 0x92, 0xDE, 0x4C, 0x80, 0x80, 0xFE, 0xFE, 0x80, 0x80, 0xFC, 0xFE, 0x02, 0x02, 0x02, 0xFE, 0xFC,
  0xF0, 0xF8, 0x0C, 0x06, 0x0C, 0xF8, 0xF0, 0xFE, 0xFE, 0x0C, 0x30, 0x0C, 0xFE, 0xFE, 0x86, 0xCE, 0x78, 0x30, 0x78, 0xCE, 0x86, 0xE0, 0xF0, 0x1E, 0x1E, 0xF0, 0xE0, 0xC6, 0xCE, 0x9A, 0x92, 0xB2,
  0xE6, 0xC6, 0x80, 0x40, 0x20, 0x10, 0x08, 0x04, 0x02, 0x82, 0x82, 0xFE, 0xFE, 0x20, 0x40, 0x80, 0x40, 0x20, 0x02, 0x02, 0x02, 0x02, 0x02, 0xC0, 0xA0, 0x04, 0x2E, 0x2A, 0x2A, 0x3E, 0x1C, 0x02,
  0xFE, 0xFE, 0x22, 0x22, 0x22, 0x3E, 0x1C, 0x3E, 0x22, 0x22, 0x22, 0x36, 0x14, 0x0C, 0x1E, 0x12, 0x12, 0x12, 0xFE, 0xFE, 0x1C, 0x3E, 0x2A, 0x2A, 0x2A, 0x3A, 0x10, 0x7E, 0xFE, 0x90, 0x80, 0x18,
  0x3D, 0x25, 0x25, 0x15, 0x3F, 0x3E, 0xFE, 0xFE, 0x20, 0x20, 0x20, 0x3E, 0x1E, 0x22, 0xBE, 0xBE, 0x02, 0x03, 0x01, 0x01, 0xBF, 0xBE, 0xFE, 0xFE, 0x08, 0x18, 0x3C, 0x26, 0x02, 0x82, 0xFE, 0xFE,
  0x02, 0x3E, 0x3E, 0x20, 0x1E, 0x20, 0x3E, 0x1E, 0x3E, 0x3E, 0x20, 0x20, 0x20, 0x3E, 0x1E, 0x1C, 0x3E, 0x22, 0x22, 0x22, 0x3E, 0x1C, 0x3F, 0x3F, 0x24, 0x24, 0x24, 0x3C, 0x18, 0x3C, 0x24, 0x24,
  0x3F, 0x3F, 0x02, 0x3E, 0x3E, 0x10, 0x30, 0x20, 0x20, 0x20, 0x10, 0x3A, 0x2A, 0x2A, 0x2A, 0x2E, 0x04, 0x20, 0xFC, 0xFE, 0x22, 0x02, 0x3C, 0x3E, 0x02, 0x02, 0x02, 0x3E, 0x3C, 0x30, 0x38, 0x0C,
  0x06, 0x0C, 0x38, 0x30, 0x3C, 0x3E, 0x02, 0x0C, 0x02, 0x3E, 0x3C, 0x22, 0x36, 0x1C, 0x08, 0x1C, 0x36, 0x22, 0x38, 0x3D, 0x05, 0x05, 0x05, 0x3F, 0x3E, 0x22, 0x26, 0x2E, 0x2A, 0x3A, 0x32, 0x22,
  0x10, 0x7C, 0xEE, 0x82, 0xEE, 0x7C, 0x10, 0x20, 0x40, 0x20, 0x40, 0x1E, 0x22, 0x42, 0x82, 0x42, 0x22, 0x1E, 0x3E, 0x68, 0xC8, 0x68, 0x3E, 0x1E, 0xFE, 0xFE, 0x92, 0x92, 0xD2, 0xDE, 0x0C, 0xFE,
  0xFE, 0x80, 0x80, 0x80, 0xC0, 0xC0, 0x02, 0x3E, 0x7E, 0x82, 0xFE, 0xFE, 0x02, 0xC6, 0xEE, 0x38, 0xFE, 0x38, 0xEE, 0xC6, 0x44, 0xC6, 0x82, 0x82, 0x92, 0xFE, 0x6C, 0xFE, 0xFE, 0x04, 0x08, 0x10,
  0xFE, 0xFE, 0x7E, 0x84, 0x48, 0x90, 0x7E, 0xFE, 0x02, 0x7E, 0xFC, 0x80, 0x80, 0xFE, 0xFE, 0x30, 0x0C, 0x30, 0xFE, 0xFE, 0x80, 0x80, 0x80, 0xFE, 0xFE, 0xF0, 0xFA, 0x0A, 0x0A, 0x0A, 0xFE, 0xFC,
  0x30, 0x48, 0xFE, 0xFE, 0x48, 0x30, 0x86, 0xCE, 0x78, 0x38, 0x78, 0xCE, 0x86, 0xFE, 0xFE, 0x02, 0x02, 0xFE, 0xFF, 0x03, 0xF0, 0xF8, 0x08, 0x08, 0x08, 0xFE, 0xFE, 0x02, 0x3E, 0x02, 0xFE, 0xFE,
  0x02, 0x3E, 0x02, 0xFE, 0xFF, 0x03, 0x80, 0xFE, 0xFE, 0x12, 0x12, 0x1E, 0x0C, 0xFE, 0xFE, 0x12, 0x1E, 0x0C, 0xF2, 0xFE, 0xFE, 0x12, 0x12, 0x12, 0x1E, 0x0C, 0x44, 0xC6, 0x82, 0x92, 0x92, 0xFE,
  0x7C, 0xFE, 0xFE, 0x10, 0x7C, 0x82, 0xFE, 0x7C, 0x70, 0xFA, 0x8E, 0x8C, 0x88, 0xFE, 0xFE, 0x1C, 0x3E, 0x72, 0x52, 0xD2, 0x9E, 0x0C, 0x3E, 0x3E, 0x2A, 0x2A, 0x3A, 0x1E, 0x04, 0x3E, 0x3E, 0x20,
  0x20, 0x30, 0x30, 0x02, 0x1E, 0x3E, 0x22, 0x3E, 0x3E, 0x02, 0x1C, 0x3E, 0x2A, 0x2A, 0x2A, 0x3A, 0x18, 0x36, 0x3E, 0x08, 0x3E, 0x08, 0x3E, 0x36, 0x14, 0x36, 0x22, 0x2A, 0x2A, 0x3E, 0x14, 0x3E,
  0x3E, 0x04, 0x08, 0x10, 0x3E, 0x3E, 0x84, 0x48, 0x90, 0x3E, 0x3E, 0x08, 0x18, 0x3C, 0x26, 0x02, 0x1E, 0x3C, 0x20, 0x20, 0x3E, 0x3E, 0x10, 0x08, 0x10, 0x3E, 0x3E, 0x08, 0x08, 0x08, 0x3E, 0x3E,
  0x20, 0x20, 0x20, 0x3E, 0x3E, 0x3F, 0x3F, 0x22, 0x22, 0x22, 0x3E, 0x1C, 0x20, 0x20, 0x3E, 0x3E, 0x20, 0x20, 0x30, 0x39, 0x0D, 0x07, 0x0E, 0x3C, 0x30, 0x08, 0x14, 0x3E, 0x3E, 0x14, 0x08, 0x3E,
  0x3E, 0x02, 0x02, 0x3E, 0x3F, 0x03, 0x30, 0x38, 0x08, 0x08, 0x08, 0x3E, 0x3E, 0x02, 0x1E, 0x02, 0x3E, 0x3E, 0x02, 0x1E, 0x02, 0x3E, 0x3F, 0x03, 0x20, 0x3E, 0x3E, 0x0A, 0x0A, 0x0E, 0x04, 0x3E,
  0x3E, 0x0A, 0x0A, 0x0E, 0x34, 0x3E, 0x3E, 0x0A, 0x0A, 0x0A, 0x0E, 0x04, 0x14, 0x36, 0x22, 0x2A, 0x2A, 0x3E, 0x1C, 0x3E, 0x3E, 0x08, 0x1C, 0x22, 0x3E, 0x1C, 0x10, 0x3A, 0x2E, 0x2C, 0x28, 0x3E,
  0x3E, 0x7E, 0xFE, 0xD2, 0x52, 0xC2, 0xE6, 0x66, 0x1C, 0xBE, 0xAA, 0x2A, 0xAA, 0xBA, 0x18
};


int8_t text_position_X = 0;
int8_t text_position_Y = 0;
PGM_VOID_P text_font = &font_normal;

#define _TEXT_UTF8_MAP_FROM 0xD0

inline uint8_t _map_UTF8_to_font(unsigned char first, unsigned char second) {
  if (first == 0xD0) {
    if ((second >= 0x90) && (second <= 0xBF)) {
      return second - 0x10;
    } else if (second == 0x81) { // Ё
      return 0xC0;
    }
  } else if (first == 0xD1) {
    if ((second >= 0x80) && (second <= 0x8F)) {
      return second + 0x30;
    } else if (second == 0x91) {
      return 0xC1; // ё
    }
  }
  return 0;
}


inline void text_set_pos(int8_t x, int8_t y) {
  text_position_X = x;
  text_position_Y = y;
}

inline void text_set_font(uint8_t code) {
  if (code == FONT_NORMAL) {
    text_font = &font_normal;
  } else if (code == FONT_BOLD) {
    text_font = &font_bold;
  } else if (code == FONT_SMALL) {
    text_font = &font_micro;
  }
}

inline void text_set_user_font(PGM_P font) {
  text_font = font;
}

inline int8_t text_get_X() {
  return text_position_X;
}

inline int8_t text_get_Y() {
  return text_position_Y;
}


uint8_t _output_char(uint8_t x, uint8_t y, uint8_t idx, PGM_VOID_P pgmFont, uint8_t fontFlags, uint8_t fontStart, uint8_t fontMax, uint8_t andMask, uint8_t xorMask)  {
  if (fontFlags & FONT_FLAG_BLANK_BEFORE) {
    display_vbits_single(x, y, andMask, xorMask, fontFlags & FONT_FLAG_MSBUP);
    x++;
    if (x >= DISPLAY_WIDTH)
      return x;
  }
  if ((idx >= fontStart) && (idx <= fontMax)) {
    uint16_t fw = pgm_read_word(pgmFont + 4 + (idx - fontStart) * 2);
    uint8_t w = (fw >> 12);
    display_vbits_pgm(x, y, pgmFont + (fw & 0xFFF), w, andMask, xorMask, fontFlags & FONT_FLAG_MSBUP);
    if (x >= DISPLAY_WIDTH - w) 
      return DISPLAY_WIDTH;
    x += w;
  }
  if (fontFlags & FONT_FLAG_BLANK_AFTER) {
    display_vbits_single(x, y, andMask, xorMask, fontFlags & FONT_FLAG_MSBUP);
    x++;
  }
  return x;
}

void text_print_char(uint8_t ch, uint8_t xorMask) {
  int8_t x = text_position_X;
  if (x >= DISPLAY_WIDTH)
  return;
  int8_t y = text_position_Y;
  uint8_t fontFlags = pgm_read_byte(text_font);
  uint8_t fontSize = pgm_read_byte(text_font + 1);
  uint8_t andMask;
  if (fontSize == 0x07) {
    andMask = 0x00;
  } else {
    andMask = (fontFlags & FONT_FLAG_MSBUP) ? shr(0x7F, fontSize & 7) : shl(0xFE, fontSize & 7);
    if (fontSize & 0x70) {
      uint8_t tt = ((fontSize >> 4) & 0x07);
      andMask |= (fontFlags & FONT_FLAG_MSBUP) ? shl(0xFF, (8 - tt)) : shr(0xFF, (8 - tt));
      y -= tt;
    }
    xorMask &= ~andMask;
  }
  if ((y < -8) || (y >= DISPLAY_HEIGHT))
    return;
  uint8_t fontStart = pgm_read_byte(text_font + 2);
  uint8_t fontMax = pgm_read_byte(text_font + 3);
  x = _output_char(x, y, ch, text_font, fontFlags, fontStart, fontMax, andMask, xorMask);
  text_position_X = x;
}


void text_print_pgm(PGM_VOID_P pgmStr, uint8_t xorMask) {
  int8_t x = text_position_X;
  if (x >= DISPLAY_WIDTH)
    return;
  int8_t y = text_position_Y;
  uint8_t fontFlags = pgm_read_byte(text_font);
  uint8_t fontSize = pgm_read_byte(text_font + 1);
  uint8_t andMask;
  if (fontSize == 0x07) {
    andMask = 0x00;
  } else {
    andMask = (fontFlags & FONT_FLAG_MSBUP) ? shr(0x7F, fontSize & 7) : shl(0xFE, fontSize & 7);
    if (fontSize & 0x70) {
      uint8_t tt = ((fontSize >> 4) & 0x07);
      andMask |= (fontFlags & FONT_FLAG_MSBUP) ? shl(0xFF, (8 - tt)) : shr(0xFF, (8 - tt));
      y -= tt;
    }
    xorMask &= ~andMask;
  }
  if ((y < -8) || (y >= DISPLAY_HEIGHT))
    return;
  uint8_t fontStart = pgm_read_byte(text_font + 2);
  uint8_t fontMax = pgm_read_byte(text_font + 3);
  do  {
    uint8_t ch = pgm_read_byte(pgmStr++);
    if (!ch) break;
    if (ch >= _TEXT_UTF8_MAP_FROM)
      ch = _map_UTF8_to_font(ch, pgm_read_byte(pgmStr++));
    x = _output_char(x, y, ch, text_font, fontFlags, fontStart, fontMax, andMask, xorMask);
  } while (x < DISPLAY_WIDTH);
  text_position_X = x;
}


void _text_print_num16(uint16_t number, uint8_t decPoint, uint8_t xorMask, uint8_t asSigned) {
  int8_t x = text_position_X;
  if (x >= DISPLAY_WIDTH)
    return;
  int8_t y = text_position_Y;
  uint8_t fontFlags = pgm_read_byte(text_font);
  uint8_t fontSize = pgm_read_byte(text_font + 1);
  uint8_t andMask;
  if (fontSize == 0x07) {
    andMask = 0x00;
  } else {
    andMask = (fontFlags & FONT_FLAG_MSBUP) ? shr(0x7F, fontSize & 7) : shl(0xFE, fontSize & 7);
    if (fontSize & 0x70) {
      uint8_t tt = ((fontSize >> 4) & 0x07);
      andMask |= (fontFlags & FONT_FLAG_MSBUP) ? shl(0xFF, (8 - tt)) : shr(0xFF, (8 - tt));
      y -= tt;
    }
    xorMask &= ~andMask;
  }
  if ((y <= -8) || (y >= DISPLAY_HEIGHT))
    return;
  uint8_t fontStart = pgm_read_byte(text_font + 2);
  uint8_t fontMax = pgm_read_byte(text_font + 3);

  uint8_t started = 0;
  if (asSigned) {
    if ((int16_t)number < 0) {
      number = -(int16_t)number;
      x = _output_char(x, y, '-', text_font, fontFlags, fontStart, fontMax, andMask, xorMask);
    }
  }
  for (int8_t i = 4; i >= 0; i--) {
    uint16_t sub;
    switch (i) {
    case 4: sub = 10000; break;
    case 3: sub = 1000; break;
    case 2: sub = 100; break;
    case 1: sub = 10; break;
    default: sub = 1;
    }
    uint8_t c = 0;
    while (number >= sub) {
      number -= sub;
      c++;
    }
    if (started || (c != 0) || (decPoint == i)) {
      x = _output_char(x, y, c + 48, text_font, fontFlags, fontStart, fontMax, andMask, xorMask);
      if (x >= DISPLAY_WIDTH)
        break;
      if ((decPoint == i) && i) {
        x = _output_char(x, y, '.', text_font, fontFlags, fontStart, fontMax, andMask, xorMask);
        if (x >= DISPLAY_WIDTH)
          break;
      }
      started = 1;
    }
    
  }
  text_position_X = x;
}

inline void text_print_num16(uint16_t number, uint8_t decPoint, uint8_t xorMask) {
  _text_print_num16(number, decPoint, xorMask, 0);
}

inline void text_print_num16s(int16_t number, uint8_t decPoint, uint8_t xorMask) {
  _text_print_num16(number, decPoint, xorMask, 1);
}

int8_t text_height() {
  uint8_t fontSize = pgm_read_byte(text_font + 1);
  return (fontSize & 0x07) - ((fontSize >> 4) & 0x07) + 1;
}

uint8_t text_width_pgm(PGM_VOID_P pgmStr) {
  uint8_t fontFlags = pgm_read_byte(text_font);
  uint8_t fontStart = pgm_read_byte(text_font + 2);
  uint8_t fontMax = pgm_read_byte(text_font + 3);
  uint8_t w = 0;
  for(;;)  {
    uint8_t ch = pgm_read_byte(pgmStr++);
    if (!ch) break;
    if (ch >= _TEXT_UTF8_MAP_FROM)
      ch = _map_UTF8_to_font(ch, pgm_read_byte(pgmStr++));
    if (fontFlags & FONT_FLAG_BLANK_BEFORE)
      w++;
    if (fontFlags & FONT_FLAG_BLANK_AFTER)
      w++;
    if ((ch >= fontStart) && (ch <= fontMax)) {
      w += pgm_read_byte(text_font + 5 + (ch - fontStart) * 2) >> 4;
    }
  };
  return w;
}

uint8_t text_width_char(uint8_t ch) {
  uint8_t fontFlags = pgm_read_byte(text_font);
  uint8_t fontStart = pgm_read_byte(text_font + 2);
  uint8_t fontMax = pgm_read_byte(text_font + 3);
  uint8_t w = 0;
  if (fontFlags & FONT_FLAG_BLANK_BEFORE)
    w++;
  if (fontFlags & FONT_FLAG_BLANK_AFTER)
    w++;
  if ((ch >= fontStart) && (ch <= fontMax)) {
    w += pgm_read_byte(text_font + 5 + (ch - fontStart) * 2) >> 4;
  }
  return w;
}

void text_print_hex(uint8_t hex, uint8_t xorMask) {
  int8_t x = text_position_X;
  if (x >= DISPLAY_WIDTH)
  return;
  int8_t y = text_position_Y;
  uint8_t fontFlags = pgm_read_byte(text_font);
  uint8_t fontSize = pgm_read_byte(text_font + 1);
  uint8_t andMask;
  if (fontSize == 0x07) {
    andMask = 0x00;
  } else {
    andMask = (fontFlags & FONT_FLAG_MSBUP) ? shr(0x7F, fontSize & 7) : shl(0xFE, fontSize & 7);
    if (fontSize & 0x70) {
      uint8_t tt = ((fontSize >> 4) & 0x07);
      andMask |= (fontFlags & FONT_FLAG_MSBUP) ? shl(0xFF, (8 - tt)) : shr(0xFF, (8 - tt));
      y -= tt;
    }
    xorMask &= ~andMask;
  }
  if ((y <= -8) || (y >= DISPLAY_HEIGHT))
    return;
  uint8_t fontStart = pgm_read_byte(text_font + 2);
  uint8_t fontMax = pgm_read_byte(text_font + 3);

  uint8_t c = (hex >> 4) + 48;
  if (c >= 58) c += 7;
  x = _output_char(x, y, c, text_font, fontFlags, fontStart, fontMax, andMask, xorMask);
  if (x < DISPLAY_WIDTH) {
    c = (hex & 0x0F) + 48;
    if (c >= 58) c += 7;
    x = _output_char(x, y, c, text_font, fontFlags, fontStart, fontMax, andMask, xorMask);
  }
  text_position_X = x;
}

